from typing import List

globals: "_Globals"

class _Globals:
    dialect_search_modules: List[str]
    def _register_dialect_impl(self, dialect_namespace: str, dialect_class: type) -> None: ...
    def _register_operation_impl(self, operation_name: str, operation_class: type, raw_opview_class: type) -> None: ...
    def append_dialect_search_prefix(self, module_name: str) -> None: ...

def register_dialect(dialect_class: type) -> object: ...
def register_operation(dialect_class: type) -> object: ...
