  .hidden archiveHidden
  .globl  archiveHidden
archiveHidden:
  .functype archiveHidden () -> (i32)
  i32.const 0
  end_function

  .globl  archiveDefault
archiveDefault:
  .functype archiveDefault () -> (i32)
  i32.const 0
  end_function
