#include <string>
#include <vector>

int main()
{
    std::vector<bool> vBool;

    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(false);
    vBool.push_back(true);
    
    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(false);
    vBool.push_back(true);

    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(false);
    vBool.push_back(true);

    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(false);
    vBool.push_back(true);

    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(false);
    vBool.push_back(true);

    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(false);
    vBool.push_back(true);
    vBool.push_back(true);

    printf ("size: %d", (int) vBool.size()); // Set break point at this line.
    return 0; 
}
